#include <stdlib.h>
#include <string.h>

#ifdef ENABLE_NLS
#include <libintl.h>
#endif

#define         _ISOC9X_SOURCE  1
#define         _ISOC99_SOURCE  1
#define         __USE_ISOC99    1
#define         __USE_ISOC9X    1

#include <math.h>
#include "utils.h"
#include "ladspa.h"

#line 30 "foo_limiter_v2.xml"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define FOO_LIMITER_MAX_ATTACK_MS        5.3f
#define FOO_LIMITER_DEFAULT_ATTACK_MS        3.0f

#define FOO_LIMITER_MAX_LOGSCALE        10.0f


typedef struct _CurvePoint {
        float gain;
        float delta;
} CurvePoint;

#ifdef FOO_TESTER
#define debug(f, s...) printf(f, ## s)
#else
#define debug(f, s...)
#endif

#define RINGBUFFER_SEEK(env, offset) ((env->ringbuffer_at+env->attack_length+(offset)) % (env->attack_length + 1))

#define RINGBUFFER_NEXT(env) RINGBUFFER_SEEK(env,  1)
#define RINGBUFFER_LAST(env) RINGBUFFER_SEEK(env, -1)

//#define RINGBUFFER_NEXT(env) (((env->ringbuffer_at)+1)                      % (env->attack_length))
//#define RINGBUFFER_LAST(env) (((env->ringbuffer_at)+(env->attack_length)-1) % (env->attack_length))

#define ENVELOPE_GAIN(env, offset) (env)->ringbuffer[ RINGBUFFER_SEEK((env), offset) ].gain

typedef struct _Envelope {
        // A ringbuffer of CurvePoints, size
        // is floor(samplerate * FOO_LIMITER_MAX_ATTACK_MS / 1000)  
        CurvePoint *ringbuffer; 
        int ringbuffer_at;

        

        // This envelope only moves forward, but "at" is needed to
        // calculate the log curve etc.
        int at;

        // attack_length also equals the amount of lookahead and
        // the used amount of the ringbuffer
        int attack_length;
        int release_length;

        // envelope_length = attack_length + release_length
        int envelope_length;

        // This is the target gain to reach when the limiter is 
        // at the precipice
        // This functions both as a target gain when ramping up
        // and it defines how the logarithmic release curve
        // is "scaled to fit"
        float limit_gain;


        // The logscale this envelope produces. Kept here to prevent
        // "zipper noise" from parameter changes while releasing
        float logscale;

} Envelope;


// The function uses the point in the ringbuffer at ( offset-1) to determine the
// current change rate and works from there.
// This function only works forward AND when the previous value has been calculated!!
static inline void FooLimiter2_calculateCurvePoint(Envelope *env, int offset, float max_delta_change)
{
        // We'll ignore rate limiting for now, let's make the basics work for us first

        int at_envelope = env->at + offset;

        CurvePoint *prev_cpoint = &env->ringbuffer[RINGBUFFER_SEEK(env, offset-1)];
        CurvePoint *this_cpoint = &env->ringbuffer[RINGBUFFER_SEEK(env, offset)];
        float delta;

        // If the envelope is over. No rate limiting done here either, boring stuff
        if (at_envelope > env->envelope_length) {
                this_cpoint->gain  = 1.0f;
                this_cpoint->delta = 0.0f;

                return;
        } 

        // Attack phase
        if (at_envelope < env->attack_length) {
                
                // Here we would do rate limiting, but we won't.. yet
                
                int samples_to_precipice = env->attack_length - at_envelope;

                delta = (env->limit_gain - prev_cpoint->gain) / (float)(samples_to_precipice+1);
                if (delta > max_delta_change) {
                        delta = max_delta_change;
                } else if (delta < -max_delta_change) {
                        delta = -max_delta_change;
                }

                this_cpoint->delta = delta;

                this_cpoint->gain  = prev_cpoint->gain + delta;

                return;
        }

        // Release phase

        // calculate the point in the relese phase we are, result value = 0.0 .. 1.0
        float pos_release = (float)( at_envelope - env->attack_length) / (float)(env->release_length);
        //debug("pos = %f\n",pos_release);


        this_cpoint->delta = (1.0f - env->limit_gain)/env->logscale/( pos_release * expf(env->logscale)+1.0f-pos_release) / (float)(env->release_length);
        this_cpoint->gain  = env->limit_gain + ( (1.0f - env->limit_gain) * (logf(pos_release * expf(env->logscale) + 1.0f - (pos_release)) / env->logscale));
}

// We need a few envelope functions:
//
// restartEnvelope() builds a completely new envelope. This one is triggered when
// The previous envelope is not running anymore or is at the release phase.
//     It takes into account the previous curve delta and ramps up the attack accordingly
//
// extendEnvelope() extends the attack current envelope (which needs to be still ramping up).
// Depending on the newly upcoming peak, the attack is either modified from this point onwards
// OR it is extended after the current peak value by a less steep curve
//
// Both functions assume that they need to reach "limit_gain" in the either parametrized or current
// attack_length samples

void FooLimiter2_restartEnvelope(Envelope *env, int attack_samples, int release_samples, float limit_gain, float logscale, float max_delta_change)
{
        int i;

        // Is the right one the previous delta, current delta or next delta?! who knows...
        CurvePoint prev_cpoint = env->ringbuffer[RINGBUFFER_LAST(env)];
        
        env->at = 0;

        env->attack_length   = attack_samples;
        debug("attack_length = %d\n",env->attack_length);
        env->release_length  = release_samples;
        env->envelope_length = attack_samples + release_samples;
        env->logscale        = 1.0f / expf(1.0f) + logscale * (FOO_LIMITER_MAX_LOGSCALE - 1.0f/expf(1.0f));

        env->limit_gain      = limit_gain;

        // We reset the ringbuffer start point.
        env->ringbuffer_at = 0;

        
        // Calculate "lookahead" (=attack_length) worth of envelope

        // We set the last position in the ringbuffer to the stored previous CurvePoint
        // so that calculateNextCurvePoint() will work
        env->ringbuffer[RINGBUFFER_LAST(env)] = prev_cpoint;

        for (i = 0; i < env->attack_length; i++) {
                FooLimiter2_calculateCurvePoint(env, i, max_delta_change);
        }
}

// Keeps the current attack time, as we are still in the attack phase. release time can be
// changed without any problems.
//
// As we are still in the attack phase, it means that there are "env->at" samples of the previously
// calculated release phase in the ringbuffer. So -env->at samples is the position of the previous
// precipice. This function divides into two different parts:
// A If the new limit_gain point requires a steeper slope than the old limit_gain
//     We start ramping up the complete attack phase with the new delta
// B If the new limit_gain point requires a shallower slope than the old limit_gain
//     We continue on the current slope, but after reaching the original limit point, we start
//     a secondary attack phase targeted to reach the newly found peak.
//     We shall call this "intelligent sustain"
//
// Note I)
// Correction regarding "steepness" of curve. In the limiter sense, the curve is "steepest" when
// its' delta is lower than the others'. That is: 
//   "Curve A is steeper than curve B, if A attenuates higher peaks than B"
// So steepness in this limiter is about curves moving faster towards -inf.
void FooLimiter2_extendEnvelope(Envelope *env, int release_samples, float new_limit_gain, float logscale, float max_delta_change)
{
        int i;
        float current_ramp_delta, new_ramp_delta;

        CurvePoint prev_cpoint = env->ringbuffer[RINGBUFFER_LAST(env)];

        env->release_length  = release_samples;
        env->envelope_length = env->attack_length + release_samples;
        env->logscale        = 1.0f / expf(1.0f) + logscale * (FOO_LIMITER_MAX_LOGSCALE - 1.0f/expf(1.0f));

        // current ramp is: (current gain - limit gain) / samples left to current precipice
        current_ramp_delta = (env->limit_gain - prev_cpoint.gain) / (float)(env->attack_length - env->at);

        // the new ramp is: (current gain - new limit gain) / attack length
        new_ramp_delta     = (new_limit_gain - prev_cpoint.gain) / (float)(env->attack_length);
        
        debug(" comparing current_ramp_delta: %f, and new_ramp_delta: %f\n",current_ramp_delta, new_ramp_delta);

        // See "Note I" about delta comparison
        if (new_ramp_delta < current_ramp_delta) {
                // Case A

                debug(" * selected case A\n");

                env->limit_gain = new_limit_gain;
                env->at = 0;

                for (i = 0; i < env->attack_length; i++) {
                        FooLimiter2_calculateCurvePoint(env, i, max_delta_change);
                }

        } else {
                // Case B, "intelligent sustain"

                i = (env->attack_length - env->at);
                env->at = 0;

#define INTELLIGENT_SUSTAIN

#ifndef INTELLIGENT_SUSTAIN

                debug(" * selected case B \"extend the current ramp\"\n");

                // We have "i" samples left.

                env->limit_gain += current_ramp_delta * (i);

#else
                // sounds good, but misses a few transients
                debug(" * selected case B \"intelligent sustain\"\n");


                // The curve up to the old precipice is already calculated,
                // If we set a new limit_gain and reset env->at, the
                // calculateNextCurvePoint() will do the "heavy lifting" as it
                // will automatically calculate the correct envelope for the
                // time between the old precipice and the new one

                // The time of the old precipice

                env->limit_gain = new_limit_gain;
#endif


                for (; i < env->attack_length; i++) {
                        FooLimiter2_calculateCurvePoint(env, i, max_delta_change);
                }

        }

}

#define FOO_LIMITER_V2_INPUT_GAIN_DB   0
#define FOO_LIMITER_V2_MAX_PEAK_DB     1
#define FOO_LIMITER_V2_ATTACK_TIME_MS  2
#define FOO_LIMITER_V2_RELEASE_TIME    3
#define FOO_LIMITER_V2_ATTENUATION     4
#define FOO_LIMITER_V2_INPUT_LEFT      5
#define FOO_LIMITER_V2_INPUT_RIGHT     6
#define FOO_LIMITER_V2_OUTPUT_LEFT     7
#define FOO_LIMITER_V2_OUTPUT_RIGHT    8
#define FOO_LIMITER_V2_LATENCY         9
#define FOO_LIMITER_V2_RELEASE_SCALE   10

static LADSPA_Descriptor *foo_limiter_v2Descriptor = NULL;

typedef struct {
	LADSPA_Data *input_gain_db;
	LADSPA_Data *max_peak_db;
	LADSPA_Data *attack_time_ms;
	LADSPA_Data *release_time;
	LADSPA_Data *attenuation;
	LADSPA_Data *input_left;
	LADSPA_Data *input_right;
	LADSPA_Data *output_left;
	LADSPA_Data *output_right;
	LADSPA_Data *latency;
	LADSPA_Data *release_scale;
	Envelope     envelope;
	int          max_attack_samples;
	float        max_delta_change;
	unsigned long samplerate;
	float *      workbuffer_left;
	float *      workbuffer_right;
	int          workbuffer_size;
	LADSPA_Data run_adding_gain;
} Foo_limiter_v2;

const LADSPA_Descriptor *ladspa_descriptor(unsigned long index) {

	switch (index) {
	case 0:
		return foo_limiter_v2Descriptor;
	default:
		return NULL;
	}
}

static void cleanupFoo_limiter_v2(LADSPA_Handle instance) {
#line 338 "foo_limiter_v2.xml"
	Foo_limiter_v2 *plugin_data = (Foo_limiter_v2 *)instance;
	free(plugin_data->workbuffer_left);
	free(plugin_data->workbuffer_right);
	free(plugin_data->envelope.ringbuffer);
	free(instance);
}

static void connectPortFoo_limiter_v2(
 LADSPA_Handle instance,
 unsigned long port,
 LADSPA_Data *data) {
	Foo_limiter_v2 *plugin;

	plugin = (Foo_limiter_v2 *)instance;
	switch (port) {
	case FOO_LIMITER_V2_INPUT_GAIN_DB:
		plugin->input_gain_db = data;
		break;
	case FOO_LIMITER_V2_MAX_PEAK_DB:
		plugin->max_peak_db = data;
		break;
	case FOO_LIMITER_V2_ATTACK_TIME_MS:
		plugin->attack_time_ms = data;
		break;
	case FOO_LIMITER_V2_RELEASE_TIME:
		plugin->release_time = data;
		break;
	case FOO_LIMITER_V2_ATTENUATION:
		plugin->attenuation = data;
		break;
	case FOO_LIMITER_V2_INPUT_LEFT:
		plugin->input_left = data;
		break;
	case FOO_LIMITER_V2_INPUT_RIGHT:
		plugin->input_right = data;
		break;
	case FOO_LIMITER_V2_OUTPUT_LEFT:
		plugin->output_left = data;
		break;
	case FOO_LIMITER_V2_OUTPUT_RIGHT:
		plugin->output_right = data;
		break;
	case FOO_LIMITER_V2_LATENCY:
		plugin->latency = data;
		break;
	case FOO_LIMITER_V2_RELEASE_SCALE:
		plugin->release_scale = data;
		break;
	}
}

static LADSPA_Handle instantiateFoo_limiter_v2(
 const LADSPA_Descriptor *descriptor,
 unsigned long s_rate) {
	Foo_limiter_v2 *plugin_data = (Foo_limiter_v2 *)malloc(sizeof(Foo_limiter_v2));
	Envelope envelope;
	int max_attack_samples;
	float max_delta_change;
	unsigned long samplerate;
	float *workbuffer_left = NULL;
	float *workbuffer_right = NULL;
	int workbuffer_size;

#line 295 "foo_limiter_v2.xml"
	samplerate = s_rate;

	max_attack_samples = (int)floor((float)samplerate * (float)FOO_LIMITER_MAX_ATTACK_MS / 1000.0f);

	/*
	//max_delta_change = 0.0001f; // TODO, calculate proper value...
	max_delta_change = 1000.0f;
	*/
	max_delta_change = 0.0002f;

	envelope.attack_length   = (int)floor((float)samplerate * FOO_LIMITER_DEFAULT_ATTACK_MS / 1000.0f);
	envelope.release_length  = 100; // just a number
	envelope.envelope_length = envelope.attack_length + envelope.release_length;
	envelope.at              = envelope.envelope_length + 1; // initial envelope is over
	envelope.limit_gain      = 0.0f; // just a number
	envelope.logscale        = 0.5f; // just another number

	envelope.ringbuffer_at   = 0;
	
	envelope.ringbuffer      = (CurvePoint *) malloc( (max_attack_samples + 1) * sizeof(CurvePoint));
	// Set the initial ringbuffer to all zeroes
	int i;
	for (i = 0; i < max_attack_samples + 1; i++) {
	        envelope.ringbuffer[i].gain  = 1.0f;
	        envelope.ringbuffer[i].delta = 0.0f;
	}


	// 4096 is just an estimate of a sane maximum buffer size. The limiter will
	// also work with bigger buffer sizes, it just has to divide processing into parts
	// (which it does internally)
	workbuffer_size  = 4096 + max_attack_samples;
	workbuffer_left  = (float *) malloc(sizeof(float) * workbuffer_size);
	workbuffer_right = (float *) malloc(sizeof(float) * workbuffer_size);

	memset(workbuffer_left,  0, sizeof(float) * workbuffer_size);
	memset(workbuffer_right, 0, sizeof(float) * workbuffer_size);

	debug("Instantiating FooLimiter v2, max_attack_samples = %d, current attack samples = %d\n",max_attack_samples, envelope.attack_length);

	plugin_data->envelope = envelope;
	plugin_data->max_attack_samples = max_attack_samples;
	plugin_data->max_delta_change = max_delta_change;
	plugin_data->samplerate = samplerate;
	plugin_data->workbuffer_left = workbuffer_left;
	plugin_data->workbuffer_right = workbuffer_right;
	plugin_data->workbuffer_size = workbuffer_size;

	return (LADSPA_Handle)plugin_data;
}

#undef buffer_write
#undef RUN_ADDING
#undef RUN_REPLACING

#define buffer_write(b, v) (b = v)
#define RUN_ADDING    0
#define RUN_REPLACING 1

static void runFoo_limiter_v2(LADSPA_Handle instance, unsigned long sample_count) {
	Foo_limiter_v2 *plugin_data = (Foo_limiter_v2 *)instance;

	/* Input gain (dB) (float value) */
	const LADSPA_Data input_gain_db = *(plugin_data->input_gain_db);

	/* Max level (dB) (float value) */
	const LADSPA_Data max_peak_db = *(plugin_data->max_peak_db);

	/* Attack time (ms) (float value) */
	const LADSPA_Data attack_time_ms = *(plugin_data->attack_time_ms);

	/* Release time (s) (float value) */
	const LADSPA_Data release_time = *(plugin_data->release_time);

	/* Input L (array of floats of length sample_count) */
	const LADSPA_Data * const input_left = plugin_data->input_left;

	/* Input R (array of floats of length sample_count) */
	const LADSPA_Data * const input_right = plugin_data->input_right;

	/* Output L (array of floats of length sample_count) */
	LADSPA_Data * const output_left = plugin_data->output_left;

	/* Output R (array of floats of length sample_count) */
	LADSPA_Data * const output_right = plugin_data->output_right;

	/* Linear/log release (float value) */
	const LADSPA_Data release_scale = *(plugin_data->release_scale);
	Envelope envelope = plugin_data->envelope;
	int max_attack_samples = plugin_data->max_attack_samples;
	float max_delta_change = plugin_data->max_delta_change;
	unsigned long samplerate = plugin_data->samplerate;
	float * workbuffer_left = plugin_data->workbuffer_left;
	float * workbuffer_right = plugin_data->workbuffer_right;
	int workbuffer_size = plugin_data->workbuffer_size;

#line 344 "foo_limiter_v2.xml"
	
	        unsigned long n,i,lookahead, buffer_offset;
	        int attack_length_parameter = (int)floor((float)samplerate * attack_time_ms / 1000.0f);
	;
	        float min_gain = 1.0f;
	
	        float peak_left, peak_right, max_peak, input_gain, current_gain;
	
	        max_peak   = DB_CO(max_peak_db);
	        input_gain = DB_CO(input_gain_db);
	
	        buffer_offset = 0;
	
	/*
	        if (envelope.at > envelope.envelope_length) {
	                envelope.attack_length = attack_length_parameter;
	                envelope.envelope_length = envelope.attack_length + envelope.release_length;
	        }
	*/
	        while (sample_count > 0) {
	            n = sample_count;
	
	            // Make sure we process in slices where "lookahead + slice" fit in the work buffer
	            if (n > workbuffer_size - max_attack_samples)
	                n = workbuffer_size - max_attack_samples;
	
	            // The start of the workbuffer contains the "latent" samples from the previous run
	            memcpy(workbuffer_left  + max_attack_samples, input_left  + buffer_offset, sizeof(float) * n);
	            memcpy(workbuffer_right + max_attack_samples, input_right + buffer_offset, sizeof(float) * n);
	
	            // _NEVER_ use the current attack length for peak detection. Use envelope.attack_length !!!!
	            // TODO: as the attack time can be modified, this is going to be tricky...
	            lookahead = envelope.attack_length;
	
	            for (i = 0; i < n; i++, lookahead++) {
	
	                // 1. input gain is applied to the work buffer
	
	                workbuffer_left [lookahead] *= input_gain;
	                workbuffer_right[lookahead] *= input_gain;
	
	                // 2. Peak detection. 
	                // Done before determining limiter gain because a new peak will have to affect
	                // the envelope at /this/ point
	
	                peak_left  = fabs(workbuffer_left [lookahead]);
	                peak_right = fabs(workbuffer_right[lookahead]);
	
	                // peak_left will contain the highest peak of the two
	                if (peak_right > peak_left) peak_left = peak_right;
	
	                float lookahead_gain = ENVELOPE_GAIN(&envelope, envelope.attack_length -1);
	
	                if ( (peak_left * lookahead_gain) > max_peak) {
	                        // a peak which isn't contained by the limiter has been found!
	
	                        int release_samples  = (int)floor( release_time * (float)samplerate);
	                        debug("Found peak (%f, max = %f) at %lu, release_samples = %d\n", peak_left, max_peak, i, release_samples);
	                        float new_limit_gain = max_peak / peak_left;
	
	
	                        // two cases, if the current envelope is in the attack phase, we call
	                        // extendEnvelope
	
	                        if (envelope.at < envelope.attack_length) {
	                                FooLimiter2_extendEnvelope(&envelope, release_samples, new_limit_gain, release_scale, max_delta_change);
	                        } else {
	                                FooLimiter2_restartEnvelope(&envelope, envelope.attack_length, release_samples, new_limit_gain, release_scale, max_delta_change);
	                                /*
	                                FooLimiter2_restartEnvelope(&envelope, attack_length_parameter, release_samples, new_limit_gain, release_scale, max_delta_change);
	
	                                // We might miss a few transients, but we don't care
	                                debug("lookahead was %lu", lookahead);
	                                lookahead = i + envelope.attack_length;
	                                debug(", lookahead is %lu\n",lookahead);
	                                */
	                        }
	                        // we have to twiddle the envelope
	        
	                }
	
	
	                
	
	                // 3. Determine and apply limiter gain
	
	                current_gain = ENVELOPE_GAIN(&envelope, 0);
	
	                if (current_gain < min_gain)
	                        min_gain = current_gain;
	
	                // 4. Brickwall limiter stage, clip whatever goes over
	
	                float out_left  = workbuffer_left[i] * current_gain;
	                float out_right = workbuffer_right[i] * current_gain;
	
	#define BRICKWALL(param, limit) if (param > limit) { param = limit; } else if (param < -limit) { param = -limit; }
	
	                BRICKWALL(out_left,  max_peak);
	                BRICKWALL(out_right, max_peak);
	
	                buffer_write(output_left [i+buffer_offset], out_left);
	
	                
	#ifndef FOO_TESTER
	                buffer_write(output_right[i+buffer_offset], out_right);
	#else
	                // when debugging, write the output envelope to the right channel
	                buffer_write(output_right[i+buffer_offset], current_gain);
	#endif
	
	
	                // 4. Advance envelope
	
	                FooLimiter2_calculateCurvePoint(&envelope,envelope.attack_length, max_delta_change);
	                if (envelope.at <= envelope.envelope_length)
	                    envelope.at++;
	                envelope.ringbuffer_at = (envelope.ringbuffer_at + 1) % (envelope.attack_length + 1);
	
	            }
	
	            // copy the "end" of input data to the lookahead buffer
	            // copy input[ (n-max_attack_samples] .. n ] => lookahead
	
	            if (n < max_attack_samples) {
	                memmove(workbuffer_left,  workbuffer_left  + n , sizeof(float) * max_attack_samples);
	                memmove(workbuffer_right, workbuffer_right + n , sizeof(float) * max_attack_samples);
	            } else {
	                memcpy (workbuffer_left,  workbuffer_left  + n , sizeof(float) * max_attack_samples);
	                memcpy (workbuffer_right, workbuffer_right + n , sizeof(float) * max_attack_samples);
	            }
	
	            sample_count -= n;
	            buffer_offset += n;
	        }
	
	        *(plugin_data->attenuation) = -CO_DB(min_gain);
	        *(plugin_data->latency)     = max_attack_samples;
	
	        plugin_data->envelope       = envelope;
}
#undef buffer_write
#undef RUN_ADDING
#undef RUN_REPLACING

#define buffer_write(b, v) (b += (v) * run_adding_gain)
#define RUN_ADDING    1
#define RUN_REPLACING 0

static void setRunAddingGainFoo_limiter_v2(LADSPA_Handle instance, LADSPA_Data gain) {
	((Foo_limiter_v2 *)instance)->run_adding_gain = gain;
}

static void runAddingFoo_limiter_v2(LADSPA_Handle instance, unsigned long sample_count) {
	Foo_limiter_v2 *plugin_data = (Foo_limiter_v2 *)instance;
	LADSPA_Data run_adding_gain = plugin_data->run_adding_gain;

	/* Input gain (dB) (float value) */
	const LADSPA_Data input_gain_db = *(plugin_data->input_gain_db);

	/* Max level (dB) (float value) */
	const LADSPA_Data max_peak_db = *(plugin_data->max_peak_db);

	/* Attack time (ms) (float value) */
	const LADSPA_Data attack_time_ms = *(plugin_data->attack_time_ms);

	/* Release time (s) (float value) */
	const LADSPA_Data release_time = *(plugin_data->release_time);

	/* Input L (array of floats of length sample_count) */
	const LADSPA_Data * const input_left = plugin_data->input_left;

	/* Input R (array of floats of length sample_count) */
	const LADSPA_Data * const input_right = plugin_data->input_right;

	/* Output L (array of floats of length sample_count) */
	LADSPA_Data * const output_left = plugin_data->output_left;

	/* Output R (array of floats of length sample_count) */
	LADSPA_Data * const output_right = plugin_data->output_right;

	/* Linear/log release (float value) */
	const LADSPA_Data release_scale = *(plugin_data->release_scale);
	Envelope envelope = plugin_data->envelope;
	int max_attack_samples = plugin_data->max_attack_samples;
	float max_delta_change = plugin_data->max_delta_change;
	unsigned long samplerate = plugin_data->samplerate;
	float * workbuffer_left = plugin_data->workbuffer_left;
	float * workbuffer_right = plugin_data->workbuffer_right;
	int workbuffer_size = plugin_data->workbuffer_size;

#line 344 "foo_limiter_v2.xml"
	
	        unsigned long n,i,lookahead, buffer_offset;
	        int attack_length_parameter = (int)floor((float)samplerate * attack_time_ms / 1000.0f);
	;
	        float min_gain = 1.0f;
	
	        float peak_left, peak_right, max_peak, input_gain, current_gain;
	
	        max_peak   = DB_CO(max_peak_db);
	        input_gain = DB_CO(input_gain_db);
	
	        buffer_offset = 0;
	
	/*
	        if (envelope.at > envelope.envelope_length) {
	                envelope.attack_length = attack_length_parameter;
	                envelope.envelope_length = envelope.attack_length + envelope.release_length;
	        }
	*/
	        while (sample_count > 0) {
	            n = sample_count;
	
	            // Make sure we process in slices where "lookahead + slice" fit in the work buffer
	            if (n > workbuffer_size - max_attack_samples)
	                n = workbuffer_size - max_attack_samples;
	
	            // The start of the workbuffer contains the "latent" samples from the previous run
	            memcpy(workbuffer_left  + max_attack_samples, input_left  + buffer_offset, sizeof(float) * n);
	            memcpy(workbuffer_right + max_attack_samples, input_right + buffer_offset, sizeof(float) * n);
	
	            // _NEVER_ use the current attack length for peak detection. Use envelope.attack_length !!!!
	            // TODO: as the attack time can be modified, this is going to be tricky...
	            lookahead = envelope.attack_length;
	
	            for (i = 0; i < n; i++, lookahead++) {
	
	                // 1. input gain is applied to the work buffer
	
	                workbuffer_left [lookahead] *= input_gain;
	                workbuffer_right[lookahead] *= input_gain;
	
	                // 2. Peak detection. 
	                // Done before determining limiter gain because a new peak will have to affect
	                // the envelope at /this/ point
	
	                peak_left  = fabs(workbuffer_left [lookahead]);
	                peak_right = fabs(workbuffer_right[lookahead]);
	
	                // peak_left will contain the highest peak of the two
	                if (peak_right > peak_left) peak_left = peak_right;
	
	                float lookahead_gain = ENVELOPE_GAIN(&envelope, envelope.attack_length -1);
	
	                if ( (peak_left * lookahead_gain) > max_peak) {
	                        // a peak which isn't contained by the limiter has been found!
	
	                        int release_samples  = (int)floor( release_time * (float)samplerate);
	                        debug("Found peak (%f, max = %f) at %lu, release_samples = %d\n", peak_left, max_peak, i, release_samples);
	                        float new_limit_gain = max_peak / peak_left;
	
	
	                        // two cases, if the current envelope is in the attack phase, we call
	                        // extendEnvelope
	
	                        if (envelope.at < envelope.attack_length) {
	                                FooLimiter2_extendEnvelope(&envelope, release_samples, new_limit_gain, release_scale, max_delta_change);
	                        } else {
	                                FooLimiter2_restartEnvelope(&envelope, envelope.attack_length, release_samples, new_limit_gain, release_scale, max_delta_change);
	                                /*
	                                FooLimiter2_restartEnvelope(&envelope, attack_length_parameter, release_samples, new_limit_gain, release_scale, max_delta_change);
	
	                                // We might miss a few transients, but we don't care
	                                debug("lookahead was %lu", lookahead);
	                                lookahead = i + envelope.attack_length;
	                                debug(", lookahead is %lu\n",lookahead);
	                                */
	                        }
	                        // we have to twiddle the envelope
	        
	                }
	
	
	                
	
	                // 3. Determine and apply limiter gain
	
	                current_gain = ENVELOPE_GAIN(&envelope, 0);
	
	                if (current_gain < min_gain)
	                        min_gain = current_gain;
	
	                // 4. Brickwall limiter stage, clip whatever goes over
	
	                float out_left  = workbuffer_left[i] * current_gain;
	                float out_right = workbuffer_right[i] * current_gain;
	
	#define BRICKWALL(param, limit) if (param > limit) { param = limit; } else if (param < -limit) { param = -limit; }
	
	                BRICKWALL(out_left,  max_peak);
	                BRICKWALL(out_right, max_peak);
	
	                buffer_write(output_left [i+buffer_offset], out_left);
	
	                
	#ifndef FOO_TESTER
	                buffer_write(output_right[i+buffer_offset], out_right);
	#else
	                // when debugging, write the output envelope to the right channel
	                buffer_write(output_right[i+buffer_offset], current_gain);
	#endif
	
	
	                // 4. Advance envelope
	
	                FooLimiter2_calculateCurvePoint(&envelope,envelope.attack_length, max_delta_change);
	                if (envelope.at <= envelope.envelope_length)
	                    envelope.at++;
	                envelope.ringbuffer_at = (envelope.ringbuffer_at + 1) % (envelope.attack_length + 1);
	
	            }
	
	            // copy the "end" of input data to the lookahead buffer
	            // copy input[ (n-max_attack_samples] .. n ] => lookahead
	
	            if (n < max_attack_samples) {
	                memmove(workbuffer_left,  workbuffer_left  + n , sizeof(float) * max_attack_samples);
	                memmove(workbuffer_right, workbuffer_right + n , sizeof(float) * max_attack_samples);
	            } else {
	                memcpy (workbuffer_left,  workbuffer_left  + n , sizeof(float) * max_attack_samples);
	                memcpy (workbuffer_right, workbuffer_right + n , sizeof(float) * max_attack_samples);
	            }
	
	            sample_count -= n;
	            buffer_offset += n;
	        }
	
	        *(plugin_data->attenuation) = -CO_DB(min_gain);
	        *(plugin_data->latency)     = max_attack_samples;
	
	        plugin_data->envelope       = envelope;
}

#ifndef FOO_TESTER
void _init() {
#else
void fake_init() {
#endif
	char **port_names;
	LADSPA_PortDescriptor *port_descriptors;
	LADSPA_PortRangeHint *port_range_hints;

#ifdef ENABLE_NLS
#define D_(s) dgettext(PACKAGE, s)
	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, PACKAGE_LOCALE_DIR);
#else
#define D_(s) (s)
#endif


	foo_limiter_v2Descriptor =
	 (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));

	if (foo_limiter_v2Descriptor) {
		foo_limiter_v2Descriptor->UniqueID = 3185;
		foo_limiter_v2Descriptor->Label = "foo_limiter_v2";
		foo_limiter_v2Descriptor->Properties =
		 LADSPA_PROPERTY_HARD_RT_CAPABLE;
		foo_limiter_v2Descriptor->Name =
		 D_("Foo Lookahead Limiter v2");
		foo_limiter_v2Descriptor->Maker =
		 "Sampo Savolainen <v2@iki.fi>";
		foo_limiter_v2Descriptor->Copyright =
		 "GPL";
		foo_limiter_v2Descriptor->PortCount = 11;

		port_descriptors = (LADSPA_PortDescriptor *)calloc(11,
		 sizeof(LADSPA_PortDescriptor));
		foo_limiter_v2Descriptor->PortDescriptors =
		 (const LADSPA_PortDescriptor *)port_descriptors;

		port_range_hints = (LADSPA_PortRangeHint *)calloc(11,
		 sizeof(LADSPA_PortRangeHint));
		foo_limiter_v2Descriptor->PortRangeHints =
		 (const LADSPA_PortRangeHint *)port_range_hints;

		port_names = (char **)calloc(11, sizeof(char*));
		foo_limiter_v2Descriptor->PortNames =
		 (const char **)port_names;

		/* Parameters for Input gain (dB) */
		port_descriptors[FOO_LIMITER_V2_INPUT_GAIN_DB] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_LIMITER_V2_INPUT_GAIN_DB] =
		 D_("Input gain (dB)");
		port_range_hints[FOO_LIMITER_V2_INPUT_GAIN_DB].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_0;
		port_range_hints[FOO_LIMITER_V2_INPUT_GAIN_DB].LowerBound = -20.0;
		port_range_hints[FOO_LIMITER_V2_INPUT_GAIN_DB].UpperBound = +10.0;

		/* Parameters for Max level (dB) */
		port_descriptors[FOO_LIMITER_V2_MAX_PEAK_DB] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_LIMITER_V2_MAX_PEAK_DB] =
		 D_("Max level (dB)");
		port_range_hints[FOO_LIMITER_V2_MAX_PEAK_DB].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_0;
		port_range_hints[FOO_LIMITER_V2_MAX_PEAK_DB].LowerBound = -30.0;
		port_range_hints[FOO_LIMITER_V2_MAX_PEAK_DB].UpperBound = +0.0;

		/* Parameters for Attack time (ms) */
		port_descriptors[FOO_LIMITER_V2_ATTACK_TIME_MS] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_LIMITER_V2_ATTACK_TIME_MS] =
		 D_("Attack time (ms)");
		port_range_hints[FOO_LIMITER_V2_ATTACK_TIME_MS].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_MIDDLE;
		port_range_hints[FOO_LIMITER_V2_ATTACK_TIME_MS].LowerBound = +1.0;
		port_range_hints[FOO_LIMITER_V2_ATTACK_TIME_MS].UpperBound = +10.0;

		/* Parameters for Release time (s) */
		port_descriptors[FOO_LIMITER_V2_RELEASE_TIME] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_LIMITER_V2_RELEASE_TIME] =
		 D_("Release time (s)");
		port_range_hints[FOO_LIMITER_V2_RELEASE_TIME].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_LOW;
		port_range_hints[FOO_LIMITER_V2_RELEASE_TIME].LowerBound = 0.01;
		port_range_hints[FOO_LIMITER_V2_RELEASE_TIME].UpperBound = 2.0;

		/* Parameters for Attenuation (dB) */
		port_descriptors[FOO_LIMITER_V2_ATTENUATION] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_LIMITER_V2_ATTENUATION] =
		 D_("Attenuation (dB)");
		port_range_hints[FOO_LIMITER_V2_ATTENUATION].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[FOO_LIMITER_V2_ATTENUATION].LowerBound = 0.0;
		port_range_hints[FOO_LIMITER_V2_ATTENUATION].UpperBound = +70.0;

		/* Parameters for Input L */
		port_descriptors[FOO_LIMITER_V2_INPUT_LEFT] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO;
		port_names[FOO_LIMITER_V2_INPUT_LEFT] =
		 D_("Input L");
		port_range_hints[FOO_LIMITER_V2_INPUT_LEFT].HintDescriptor = 0;

		/* Parameters for Input R */
		port_descriptors[FOO_LIMITER_V2_INPUT_RIGHT] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO;
		port_names[FOO_LIMITER_V2_INPUT_RIGHT] =
		 D_("Input R");
		port_range_hints[FOO_LIMITER_V2_INPUT_RIGHT].HintDescriptor = 0;

		/* Parameters for Output L */
		port_descriptors[FOO_LIMITER_V2_OUTPUT_LEFT] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO;
		port_names[FOO_LIMITER_V2_OUTPUT_LEFT] =
		 D_("Output L");
		port_range_hints[FOO_LIMITER_V2_OUTPUT_LEFT].HintDescriptor = 0;

		/* Parameters for Output R */
		port_descriptors[FOO_LIMITER_V2_OUTPUT_RIGHT] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO;
		port_names[FOO_LIMITER_V2_OUTPUT_RIGHT] =
		 D_("Output R");
		port_range_hints[FOO_LIMITER_V2_OUTPUT_RIGHT].HintDescriptor = 0;

		/* Parameters for latency */
		port_descriptors[FOO_LIMITER_V2_LATENCY] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_LIMITER_V2_LATENCY] =
		 D_("latency");
		port_range_hints[FOO_LIMITER_V2_LATENCY].HintDescriptor = 0;

		/* Parameters for Linear/log release */
		port_descriptors[FOO_LIMITER_V2_RELEASE_SCALE] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_LIMITER_V2_RELEASE_SCALE] =
		 D_("Linear/log release");
		port_range_hints[FOO_LIMITER_V2_RELEASE_SCALE].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_HIGH;
		port_range_hints[FOO_LIMITER_V2_RELEASE_SCALE].LowerBound = 0.0;
		port_range_hints[FOO_LIMITER_V2_RELEASE_SCALE].UpperBound = 1.0;

		foo_limiter_v2Descriptor->activate = NULL;
		foo_limiter_v2Descriptor->cleanup = cleanupFoo_limiter_v2;
		foo_limiter_v2Descriptor->connect_port = connectPortFoo_limiter_v2;
		foo_limiter_v2Descriptor->deactivate = NULL;
		foo_limiter_v2Descriptor->instantiate = instantiateFoo_limiter_v2;
		foo_limiter_v2Descriptor->run = runFoo_limiter_v2;
		foo_limiter_v2Descriptor->run_adding = runAddingFoo_limiter_v2;
		foo_limiter_v2Descriptor->set_run_adding_gain = setRunAddingGainFoo_limiter_v2;
	}
}

#ifndef FOO_TESTER
void _fini() {
#else
void fake_fini() {
#endif
	if (foo_limiter_v2Descriptor) {
		free((LADSPA_PortDescriptor *)foo_limiter_v2Descriptor->PortDescriptors);
		free((char **)foo_limiter_v2Descriptor->PortNames);
		free((LADSPA_PortRangeHint *)foo_limiter_v2Descriptor->PortRangeHints);
		free(foo_limiter_v2Descriptor);
	}

}
