#include <stdlib.h>
#include <string.h>

#ifdef ENABLE_NLS
#include <libintl.h>
#endif

#define         _ISOC9X_SOURCE  1
#define         _ISOC99_SOURCE  1
#define         __USE_ISOC99    1
#define         __USE_ISOC9X    1

#include <math.h>
#include "utils.h"
#include "ladspa.h"

#line 29 "foo_saturator.xml"

#define PARAMETER_SLOPE_SAMPLES_SECONDS 0.0025

typedef struct SaturatorParameters_
{
        float gain;
        float peak;
        float depth;
        float drive;

        float threshold; // not parametrized
} SaturatorParameters;

#define FOO_SATURATOR_INPUT_GAIN_DB    0
#define FOO_SATURATOR_MAX_PEAK_DB      1
#define FOO_SATURATOR_SQUARE           2
#define FOO_SATURATOR_DRIVE            3
#define FOO_SATURATOR_INPUT_BUFFER     4
#define FOO_SATURATOR_OUTPUT_BUFFER    5
#define FOO_SATURATOR_LATENCY          6

static LADSPA_Descriptor *foo_saturatorDescriptor = NULL;

typedef struct {
	LADSPA_Data *input_gain_db;
	LADSPA_Data *max_peak_db;
	LADSPA_Data *square;
	LADSPA_Data *drive;
	LADSPA_Data *input_buffer;
	LADSPA_Data *output_buffer;
	LADSPA_Data *latency;
	SaturatorParameters last_parameters;
	float        samplerate;
	LADSPA_Data run_adding_gain;
} Foo_saturator;

const LADSPA_Descriptor *ladspa_descriptor(unsigned long index) {

	switch (index) {
	case 0:
		return foo_saturatorDescriptor;
	default:
		return NULL;
	}
}

static void cleanupFoo_saturator(LADSPA_Handle instance) {
	free(instance);
}

static void connectPortFoo_saturator(
 LADSPA_Handle instance,
 unsigned long port,
 LADSPA_Data *data) {
	Foo_saturator *plugin;

	plugin = (Foo_saturator *)instance;
	switch (port) {
	case FOO_SATURATOR_INPUT_GAIN_DB:
		plugin->input_gain_db = data;
		break;
	case FOO_SATURATOR_MAX_PEAK_DB:
		plugin->max_peak_db = data;
		break;
	case FOO_SATURATOR_SQUARE:
		plugin->square = data;
		break;
	case FOO_SATURATOR_DRIVE:
		plugin->drive = data;
		break;
	case FOO_SATURATOR_INPUT_BUFFER:
		plugin->input_buffer = data;
		break;
	case FOO_SATURATOR_OUTPUT_BUFFER:
		plugin->output_buffer = data;
		break;
	case FOO_SATURATOR_LATENCY:
		plugin->latency = data;
		break;
	}
}

static LADSPA_Handle instantiateFoo_saturator(
 const LADSPA_Descriptor *descriptor,
 unsigned long s_rate) {
	Foo_saturator *plugin_data = (Foo_saturator *)malloc(sizeof(Foo_saturator));
	SaturatorParameters last_parameters;
	float samplerate;

#line 50 "foo_saturator.xml"
	samplerate = (float)s_rate;

	//printf("instantiate saturator\n");

	last_parameters.gain = 1.0;
	last_parameters.peak = 1.0;
	last_parameters.depth = 0.3;
	last_parameters.drive = 1.0;
	
	last_parameters.threshold = 1.0 - 0.3;

	plugin_data->last_parameters = last_parameters;
	plugin_data->samplerate = samplerate;

	return (LADSPA_Handle)plugin_data;
}

#undef buffer_write
#undef RUN_ADDING
#undef RUN_REPLACING

#define buffer_write(b, v) (b = v)
#define RUN_ADDING    0
#define RUN_REPLACING 1

static void runFoo_saturator(LADSPA_Handle instance, unsigned long sample_count) {
	Foo_saturator *plugin_data = (Foo_saturator *)instance;

	/* Input gain (dB) (float value) */
	const LADSPA_Data input_gain_db = *(plugin_data->input_gain_db);

	/* Max level (dB) (float value) */
	const LADSPA_Data max_peak_db = *(plugin_data->max_peak_db);

	/* Square-waveness (float value) */
	const LADSPA_Data square = *(plugin_data->square);

	/* Drive (float value) */
	const LADSPA_Data drive = *(plugin_data->drive);

	/* Input (array of floats of length sample_count) */
	const LADSPA_Data * const input_buffer = plugin_data->input_buffer;

	/* Output (array of floats of length sample_count) */
	LADSPA_Data * const output_buffer = plugin_data->output_buffer;
	SaturatorParameters last_parameters = plugin_data->last_parameters;
	float samplerate = plugin_data->samplerate;

#line 74 "foo_saturator.xml"
	SaturatorParameters p, new_parameters;
	int slope_length, i;
	float t, input, sign, CV = 1.0f;

	new_parameters.gain  = DB_CO(input_gain_db); // TODO: db -> coef conversion
	new_parameters.peak  = DB_CO(max_peak_db);   // ditto
	new_parameters.depth = (1.0f - square) * new_parameters.peak;
	new_parameters.drive = drive;
	new_parameters.threshold = new_parameters.peak - new_parameters.depth;

	*(plugin_data->latency) = 0;
	//latency = 0;

	//printf("Bo: gain %f, peak %f, depth %f, drive %f\n",new_parameters.gain, new_parameters.peak, new_parameters.depth, new_parameters.drive);

	/*
	printf("gain %f - %f, peak %f - %f, depth %f - %f, drive %f - %f\n",
	      last_parameters.gain,  new_parameters.gain,
	      last_parameters.peak,  new_parameters.peak,
	      last_parameters.depth, new_parameters.depth,
	      last_parameters.drive, new_parameters.drive);
	*/

	// threshold is calculated, so no need to check it
	if (( last_parameters.gain  == new_parameters.gain  &&
	      last_parameters.peak  == new_parameters.peak  &&
	      last_parameters.depth == new_parameters.depth &&
	      last_parameters.drive == new_parameters.drive) ) {
	        slope_length = 0;
	} else {
	        slope_length = samplerate * PARAMETER_SLOPE_SAMPLES_SECONDS;
	        if ( (sample_count-1) < slope_length) {
	                slope_length = sample_count-1;
	        }
	}

	float ratio;

	/*
	if (slope_length > 0) {
	        printf("Changing parameters %d %d\n", sample_count, slope_length);
	}
	*/

	for (i = 0; i < sample_count; i++) {
	        if (i < slope_length) {
	                ratio = (float)(i+1)/(float)(slope_length+1);
	                p.gain  = (1.0 - ratio) * last_parameters.gain  + ratio * new_parameters.gain;
	                p.peak  = (1.0 - ratio) * last_parameters.peak  + ratio * new_parameters.peak;
	                p.depth = (1.0 - ratio) * last_parameters.depth + ratio * new_parameters.depth;
	                p.drive = (1.0 - ratio) * last_parameters.drive + ratio * new_parameters.drive;
	                
	                p.threshold = p.peak - p.depth;
	        } else if (i == slope_length) {
	                p = new_parameters;
	                p.threshold = p.peak - p.depth;
	        }


	        input = input_buffer[i] * p.gain;
	        sign = (input < 0.0 ? -1.0 : 1.0);
	        t = fabsf(input);
	        
	        // detector
	        if (t <= p.threshold) {
	                CV = 1.0f;
	        } else {
	                // CV
	                
	                // remove the threshold
	                t -= p.threshold;

	                // scale to the area
	                t /= p.depth;

	                // apply drive
	                t *= p.drive;

	                // tan it
	                t  = tanh(t);

	                // create target sample
	                t = (p.threshold + t*p.depth) * sign; // scale it back
	                
	                // targetCV = the gain required to apply to input to reach t
	                CV = t / input;
	        }

	        buffer_write(output_buffer[i], input_buffer[i] * p.gain * CV);

	}

	plugin_data->last_parameters = new_parameters;
}
#undef buffer_write
#undef RUN_ADDING
#undef RUN_REPLACING

#define buffer_write(b, v) (b += (v) * run_adding_gain)
#define RUN_ADDING    1
#define RUN_REPLACING 0

static void setRunAddingGainFoo_saturator(LADSPA_Handle instance, LADSPA_Data gain) {
	((Foo_saturator *)instance)->run_adding_gain = gain;
}

static void runAddingFoo_saturator(LADSPA_Handle instance, unsigned long sample_count) {
	Foo_saturator *plugin_data = (Foo_saturator *)instance;
	LADSPA_Data run_adding_gain = plugin_data->run_adding_gain;

	/* Input gain (dB) (float value) */
	const LADSPA_Data input_gain_db = *(plugin_data->input_gain_db);

	/* Max level (dB) (float value) */
	const LADSPA_Data max_peak_db = *(plugin_data->max_peak_db);

	/* Square-waveness (float value) */
	const LADSPA_Data square = *(plugin_data->square);

	/* Drive (float value) */
	const LADSPA_Data drive = *(plugin_data->drive);

	/* Input (array of floats of length sample_count) */
	const LADSPA_Data * const input_buffer = plugin_data->input_buffer;

	/* Output (array of floats of length sample_count) */
	LADSPA_Data * const output_buffer = plugin_data->output_buffer;
	SaturatorParameters last_parameters = plugin_data->last_parameters;
	float samplerate = plugin_data->samplerate;

#line 74 "foo_saturator.xml"
	SaturatorParameters p, new_parameters;
	int slope_length, i;
	float t, input, sign, CV = 1.0f;

	new_parameters.gain  = DB_CO(input_gain_db); // TODO: db -> coef conversion
	new_parameters.peak  = DB_CO(max_peak_db);   // ditto
	new_parameters.depth = (1.0f - square) * new_parameters.peak;
	new_parameters.drive = drive;
	new_parameters.threshold = new_parameters.peak - new_parameters.depth;

	*(plugin_data->latency) = 0;
	//latency = 0;

	//printf("Bo: gain %f, peak %f, depth %f, drive %f\n",new_parameters.gain, new_parameters.peak, new_parameters.depth, new_parameters.drive);

	/*
	printf("gain %f - %f, peak %f - %f, depth %f - %f, drive %f - %f\n",
	      last_parameters.gain,  new_parameters.gain,
	      last_parameters.peak,  new_parameters.peak,
	      last_parameters.depth, new_parameters.depth,
	      last_parameters.drive, new_parameters.drive);
	*/

	// threshold is calculated, so no need to check it
	if (( last_parameters.gain  == new_parameters.gain  &&
	      last_parameters.peak  == new_parameters.peak  &&
	      last_parameters.depth == new_parameters.depth &&
	      last_parameters.drive == new_parameters.drive) ) {
	        slope_length = 0;
	} else {
	        slope_length = samplerate * PARAMETER_SLOPE_SAMPLES_SECONDS;
	        if ( (sample_count-1) < slope_length) {
	                slope_length = sample_count-1;
	        }
	}

	float ratio;

	/*
	if (slope_length > 0) {
	        printf("Changing parameters %d %d\n", sample_count, slope_length);
	}
	*/

	for (i = 0; i < sample_count; i++) {
	        if (i < slope_length) {
	                ratio = (float)(i+1)/(float)(slope_length+1);
	                p.gain  = (1.0 - ratio) * last_parameters.gain  + ratio * new_parameters.gain;
	                p.peak  = (1.0 - ratio) * last_parameters.peak  + ratio * new_parameters.peak;
	                p.depth = (1.0 - ratio) * last_parameters.depth + ratio * new_parameters.depth;
	                p.drive = (1.0 - ratio) * last_parameters.drive + ratio * new_parameters.drive;
	                
	                p.threshold = p.peak - p.depth;
	        } else if (i == slope_length) {
	                p = new_parameters;
	                p.threshold = p.peak - p.depth;
	        }


	        input = input_buffer[i] * p.gain;
	        sign = (input < 0.0 ? -1.0 : 1.0);
	        t = fabsf(input);
	        
	        // detector
	        if (t <= p.threshold) {
	                CV = 1.0f;
	        } else {
	                // CV
	                
	                // remove the threshold
	                t -= p.threshold;

	                // scale to the area
	                t /= p.depth;

	                // apply drive
	                t *= p.drive;

	                // tan it
	                t  = tanh(t);

	                // create target sample
	                t = (p.threshold + t*p.depth) * sign; // scale it back
	                
	                // targetCV = the gain required to apply to input to reach t
	                CV = t / input;
	        }

	        buffer_write(output_buffer[i], input_buffer[i] * p.gain * CV);

	}

	plugin_data->last_parameters = new_parameters;
}

#ifndef FOO_TESTER
void _init() {
#else
void fake_init() {
#endif
	char **port_names;
	LADSPA_PortDescriptor *port_descriptors;
	LADSPA_PortRangeHint *port_range_hints;

#ifdef ENABLE_NLS
#define D_(s) dgettext(PACKAGE, s)
	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, PACKAGE_LOCALE_DIR);
#else
#define D_(s) (s)
#endif


	foo_saturatorDescriptor =
	 (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));

	if (foo_saturatorDescriptor) {
		foo_saturatorDescriptor->UniqueID = 3187;
		foo_saturatorDescriptor->Label = "foo_saturator";
		foo_saturatorDescriptor->Properties =
		 LADSPA_PROPERTY_HARD_RT_CAPABLE;
		foo_saturatorDescriptor->Name =
		 D_("Foo Saturator");
		foo_saturatorDescriptor->Maker =
		 "Sampo Savolainen <v2@iki.fi>";
		foo_saturatorDescriptor->Copyright =
		 "GPL";
		foo_saturatorDescriptor->PortCount = 7;

		port_descriptors = (LADSPA_PortDescriptor *)calloc(7,
		 sizeof(LADSPA_PortDescriptor));
		foo_saturatorDescriptor->PortDescriptors =
		 (const LADSPA_PortDescriptor *)port_descriptors;

		port_range_hints = (LADSPA_PortRangeHint *)calloc(7,
		 sizeof(LADSPA_PortRangeHint));
		foo_saturatorDescriptor->PortRangeHints =
		 (const LADSPA_PortRangeHint *)port_range_hints;

		port_names = (char **)calloc(7, sizeof(char*));
		foo_saturatorDescriptor->PortNames =
		 (const char **)port_names;

		/* Parameters for Input gain (dB) */
		port_descriptors[FOO_SATURATOR_INPUT_GAIN_DB] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_SATURATOR_INPUT_GAIN_DB] =
		 D_("Input gain (dB)");
		port_range_hints[FOO_SATURATOR_INPUT_GAIN_DB].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_0;
		port_range_hints[FOO_SATURATOR_INPUT_GAIN_DB].LowerBound = -20.0;
		port_range_hints[FOO_SATURATOR_INPUT_GAIN_DB].UpperBound = +20.0;

		/* Parameters for Max level (dB) */
		port_descriptors[FOO_SATURATOR_MAX_PEAK_DB] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_SATURATOR_MAX_PEAK_DB] =
		 D_("Max level (dB)");
		port_range_hints[FOO_SATURATOR_MAX_PEAK_DB].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_MAXIMUM;
		port_range_hints[FOO_SATURATOR_MAX_PEAK_DB].LowerBound = -60.0;
		port_range_hints[FOO_SATURATOR_MAX_PEAK_DB].UpperBound = +0.0;

		/* Parameters for Square-waveness */
		port_descriptors[FOO_SATURATOR_SQUARE] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_SATURATOR_SQUARE] =
		 D_("Square-waveness");
		port_range_hints[FOO_SATURATOR_SQUARE].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_MIDDLE;
		port_range_hints[FOO_SATURATOR_SQUARE].LowerBound = 0.0;
		port_range_hints[FOO_SATURATOR_SQUARE].UpperBound = +1.0;

		/* Parameters for Drive */
		port_descriptors[FOO_SATURATOR_DRIVE] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_SATURATOR_DRIVE] =
		 D_("Drive");
		port_range_hints[FOO_SATURATOR_DRIVE].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_MINIMUM;
		port_range_hints[FOO_SATURATOR_DRIVE].LowerBound = 1.0;
		port_range_hints[FOO_SATURATOR_DRIVE].UpperBound = 6.0;

		/* Parameters for Input */
		port_descriptors[FOO_SATURATOR_INPUT_BUFFER] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO;
		port_names[FOO_SATURATOR_INPUT_BUFFER] =
		 D_("Input");
		port_range_hints[FOO_SATURATOR_INPUT_BUFFER].HintDescriptor = 0;

		/* Parameters for Output */
		port_descriptors[FOO_SATURATOR_OUTPUT_BUFFER] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO;
		port_names[FOO_SATURATOR_OUTPUT_BUFFER] =
		 D_("Output");
		port_range_hints[FOO_SATURATOR_OUTPUT_BUFFER].HintDescriptor = 0;

		/* Parameters for latency */
		port_descriptors[FOO_SATURATOR_LATENCY] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_SATURATOR_LATENCY] =
		 D_("latency");
		port_range_hints[FOO_SATURATOR_LATENCY].HintDescriptor = 0;

		foo_saturatorDescriptor->activate = NULL;
		foo_saturatorDescriptor->cleanup = cleanupFoo_saturator;
		foo_saturatorDescriptor->connect_port = connectPortFoo_saturator;
		foo_saturatorDescriptor->deactivate = NULL;
		foo_saturatorDescriptor->instantiate = instantiateFoo_saturator;
		foo_saturatorDescriptor->run = runFoo_saturator;
		foo_saturatorDescriptor->run_adding = runAddingFoo_saturator;
		foo_saturatorDescriptor->set_run_adding_gain = setRunAddingGainFoo_saturator;
	}
}

#ifndef FOO_TESTER
void _fini() {
#else
void fake_fini() {
#endif
	if (foo_saturatorDescriptor) {
		free((LADSPA_PortDescriptor *)foo_saturatorDescriptor->PortDescriptors);
		free((char **)foo_saturatorDescriptor->PortNames);
		free((LADSPA_PortRangeHint *)foo_saturatorDescriptor->PortRangeHints);
		free(foo_saturatorDescriptor);
	}

}
